#include "mbping_kbd.h"
#include "mbping.h"

#include "mbdef.h"

#include "mb_api.h"

#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <errno.h>

void kbd_bind( int argc, char **argv ) {
	mbping_bind( argv[ 1 ] );
}

void kbd_unbind( int argc, char **argv ) {
	mbping_unbind( );
}

void kbd_register_as( int argc, char **argv ) {
	switch ( argc ) {
	case 3:
		mbping_register_as( argv[ 1 ], 1 );
		break;
	case 2:
		mbping_register_as( argv[ 1 ], 0 );
		break;
	default:
		fprintf( stderr, "%s app_name\n", argv[ 0 ] );
		break;
	}
}

void kbd_probe( int argc, char **argv ) {
	switch ( argc ) {
	case 3:
		mbping_probe( argv[ 1 ], 1 );
		break;
	case 2:
		mbping_probe( argv[ 1 ], 0 );
		break;
	default:
		fprintf( stderr, "%s app_name\n", argv[ 0 ] );
		break;
	}
}

static void kbd_send_msg( int argc, char **argv, int how ) {
	if ( argc > 2 ) {
		u_char *msg, *p;
		u_int len = 0;

		int i;

		for( i = 2; i < argc; i++ )
			len += strlen( argv[ i ] ) + 1;

		msg = (u_char *)malloc( len + 1 );

		for( p = msg, i = 2; i < argc; i++ ) {
			strcpy( (char *)p, argv[ i ] );
			p += strlen( argv [ i ] );
			if ( i < argc - 1 )
				*p++ = ' ';
		}
		mbping_send_data( argv[ 1 ], len, msg, how );

		free( msg );
	}
	else
		fprintf( stderr, "%s app_name msg\n", argv[ 0 ] );
}

void kbd_oneway( int argc, char **argv ) {
	kbd_send_msg( argc, argv, MB_ONEWAY );
}

void kbd_twoway( int argc, char **argv ) {
	kbd_send_msg( argc, argv, MB_TWOWAY );
}

static void kbd_send_data( int argc, char **argv, int how ) {
	if ( argc > 3 ) {
		int len = atoi( argv[ 2 ] );
		if ( len > 0 ) {
			u_char *msg = (u_char *)malloc( len );
			if ( !msg )
				fprintf( stderr, "%s\n", strerror( errno ) );
			else {
				memset( msg, argv[ 3 ][ 0 ], len );

				mbping_send_data( argv[ 1 ], len, msg, how );
				free( msg );
			}
		}
	}
	else
		fprintf( stderr, "%s app_name count char\n", argv[ 0 ] );
}

void kbd_send( int argc, char **argv ) {
	kbd_send_data( argc, argv, MB_ONEWAY );
}

void kbd_multiway( int argc, char **argv ) {
	kbd_send_data( argc, argv, MB_MULTIWAY );
}

void kbd_pagesize( int argc, char **argv ) {
	switch ( argc ) {
	case 2: {
		int size = atoi( argv[ 1 ] );
		if ( size > 0 )
			mb_setpagesize( size );
		else
			goto error;
		break;
	}
	case 1:
		fprintf( stdout, "%u\n", mb_getpagesize() );
		break;
	default:
	error:
		fprintf( stderr, "%s [n_octets]\n", argv[ 0 ] );
		break;
	}
}

void kbd_dumpsize( int argc, char **argv ) {
	switch ( argc ) {
	case 2: {
		int size = atoi( argv[ 1 ] );
		if ( size > 0 )
			mbping.dump_size = size;
		else
			goto error;
		break;
	}
	case 1:
		fprintf( stdout, "%u\n", mbping.dump_size );
		break;
	default:
	error:
		fprintf( stderr, "%s [n_octets]\n", argv[ 0 ] );
		break;
	}
}

void kbd_sleep( int argc, char **argv ) {
	switch ( argc ) {
	case 2:
		sleep( atoi( argv[ 1 ] ));
		break;
	default:
		fprintf( stderr, "%s nsecs\n", argv[ 0 ] );
		break;
	}
}

void kbd_exit( int argc, char **argv ) {
	mbping_exit( 0 );
}

